"""
The LipdToRDF class helps in converting a LiPD file to an RDF Graph.
It uses the SCHEMA dictionary (from globals/schema.py) to do the conversion
"""

import copy
import json
import pickle
import re
import os
import os.path
import zipfile
import tempfile
import pandas as pd

from rdflib import RDFS
from rdflib.graph import ConjunctiveGraph, Literal, RDF, URIRef, BNode, Collection
from rdflib.namespace import XSD

from io import BytesIO
from urllib.request import urlopen
from urllib.parse import urlparse, urlunparse, quote

from ..globals.urls import NSURL, DATAURL, ONTONS, NAMESPACES
from ..globals.blacklist import BLACKLIST
from ..globals.synonyms import SYNONYMS
from ..globals.schema import SCHEMA

from .utils import expand_schema, ucfirst, lcfirst, camelCase, unCamelCase, escape, uniqid, sanitizeId

class LipdToRDF:
    """
    The LipdToRDF class helps in converting a LiPD file to an RDF Graph. 
    It uses the SCHEMA dictionary (from globals/schema.py) to do the conversion
    """

    def __init__(self, standardize=True, add_labels=True):
        self.graph = ConjunctiveGraph()
        self.lipd_csvs = {}
        self.graphurl = NSURL
        self.namespace = NSURL + "/"
        self.standardize = standardize
        self.add_labels = add_labels
        self.schema = expand_schema(copy.deepcopy(SCHEMA))  


    def convert(self, lipdpath):
        '''Convert LiPD file to RDF Graph

        Parameters
        ----------

        lipdpath : str
            path to lipd file (the path could also be a url)

        '''
        self.graph = ConjunctiveGraph()
        
        lpdname = os.path.basename(lipdpath).replace(".lpd", "")
        lpdname = re.sub(r"\?.+$", "", lpdname)

        self.graphurl = NSURL + "/" + lpdname

        with tempfile.TemporaryDirectory(prefix="lipd_to_rdf_") as tmpdir:
            self._unzip_lipd_file(lipdpath, tmpdir)
            jsons = self._find_files_with_extension(tmpdir, 'jsonld')
            for jsonpath, _ in jsons:
                jsondir = os.path.dirname(jsonpath)
                csvs = self._find_files_with_extension(jsondir, 'csv')
                self.lipd_csvs = {}
                for csvpath, _ in csvs:
                    csvname = os.path.basename(csvpath)        
                    try:
                        self.lipd_csvs[csvname] = pd.read_csv(csvpath, header=None)
                    except:
                        # If normal load doesn't work, try to detect the number of columns and load it that way
                        print(f"WARNING: CSV file '{csvname}' might have inconsistent number of columns !!\nDetecting number of columns to load ..\n")
                        self.lipd_csvs[csvname] = self._detect_columns_and_load(csvpath)
                self._load_lipd_json_to_graph(jsonpath)

    def _detect_columns_and_load(self, filename):
        # detect number of columns
        num_columns=0
        with open(filename) as f:
            for line in f.readlines():
                num = len(line.split(','))
                if num > num_columns:
                    num_columns=num
        # load with pre-determined number of columns
        df=pd.read_csv(filename,names=range(num_columns))
        return df


    def serialize(self, topath, type="rdf"):
        '''Write LiPD RDF Graph to RDF file (or Pickle file)

        Parameters
        ----------

        topath : str
            path to the output file

        type : str
            the output file type : rdf or pickle (we store the pickled rdf graph for efficiency sometimes)
        '''

        if self.graph:
            if type == "rdf":
                self.graph.serialize(topath, format="nquads", encoding="utf-8")
            elif type == "pickle":
                with open(topath, 'wb') as f:
                    pickle.dump(self.graph, f)


    def _unzip_lipd_file(self, lipdfile, unzipdir):
        if lipdfile.startswith("http"):
            # If this is a URL
            # Handle special characters in url (if any)
            res = urlparse(lipdfile)
            lipdurl = urlunparse(res._replace(path=quote(res.path)))

            # Open url and unzip
            resp = urlopen(lipdurl)
            with zipfile.ZipFile(BytesIO(resp.read())) as zip_ref:
                zip_ref.extractall(unzipdir)
        else:
            # If this is a local file
            # Unzip file
            with zipfile.ZipFile(lipdfile, 'r') as zip_ref:
                zip_ref.extractall(unzipdir)

    def _parse_persons_string(self, author_string, parent = None) :
        # Check for semi-colon delimiter and split accordingly
        if ";" in author_string:
            author_split = re.split(r"\s*;\s*", author_string)
            # Further split the authors with commas if necessary
            author_list = []
            for author in author_split:
                author = author
                if "," in author:
                    last_first = re.split(r"\s*,\s*", author)
                    author_list.append(f"{last_first[1]} {last_first[0]}")
                else:
                    author_list.append(author)
            authors = author_list
        else:
            # Split the author string with commas
            author_list = []
            author_split = re.split(r"\s*,\s*", author_string)
            if len(author_split) % 2 == 0:
                # Even number : last name first name
                for i in range(0, len(author_split), 2):
                    author_list.append(f"{author_split[i+1]} {author_split[i]}")
            else:
                # Odd number : first name last name
                for author in author_split:
                    author_list.append(author)
            authors = author_list
        return authors


    def _parse_persons(self, auths, parent = None) :
        authors = []
        if (not type(auths) is list) :
            auths = [auths]
        
        for authstr in auths: 
            authname = None
            if type(authstr) is dict:
                if "name" in authstr:
                    authname = authstr["name"]
            else:
                authname = authstr
            
            if authname:
                auth = self._parse_persons_string(authname, parent)
                if type(auth) is list:
                    authors.extend(auth)
                else:
                    authors.append(auth)
        
        return [{"name": auth} for auth in authors]


    def _parse_location(self, geo, parent = None) :
        ngeo = {}
        ngeo["locationType"] = geo["type"] if "type" in geo else None
        ngeo["coordinatesFor"] = parent["@id"]
        coords = geo["geometry"]["coordinates"]
        if (coords and len(coords) > 0) :
            ngeo["coordinates"] = str(str(coords[1]) + ",") + str(coords[0])
            ngeo["wgs84:lat"] = coords[1]
            ngeo["hasLatitude"] = coords[1]
            ngeo["latitude"] = coords[1]

            ngeo["wgs84:long"] = coords[0]
            ngeo["longitude"] = coords[0]
            ngeo["hasLongitude"] = coords[0]
            
            if (len(coords) > 2) :
                ngeo["wgs84:alt"] = coords[2]
                ngeo["elevation"] = coords[2]
                ngeo["hasElevation"] = coords[2]
        
        if "properties" in geo and isinstance(geo["properties"], dict) :
            for key,value in geo["properties"].items() :
                ngeo[key] = value
        elif isinstance(geo, dict) :
            for key,value in geo.items() :
                if key != "geometry":
                    # Do not add lat long if they are already added
                    if f"wgs84:{key}" not in ngeo:
                        ngeo[key] = value
        return ngeo


    def _get_uncertainty(self, val, parent = None) :
        uncertainty = {}
        uncertainty["hasValue"] = val
        uncertainty["analytical"] = val
        uncertainty["reproducibility"] = val
        return uncertainty

    def _get_google_spreadsheet_url(self, key, parent = None) :
        return "https://docs.google.com/spreadsheets/d/" + str(key) + ""
    
    def _get_parent_property(self, obj, prop) :
        parent = obj["@parent"]
        while (parent) :
            if ((prop in parent)) :
                return parent[prop]
            
            parent = parent["@parent"]
        return None

    def _get_parent_with_property_value(self, obj, prop, val) :
        parent = obj["@parent"]
        while (parent) :
            if ((prop in parent) and parent[prop] == val) :
                return parent
            
            parent = parent["@parent"]
        return None
    
    def _set_identifier_properties(self, pub, objhash) :
        props = {}
        if "identifier" in pub :
            for identifier in pub["identifier"] : 
                if identifier["type"] == "doi" :
                    if "hasDOI" not in pub:
                        pub["hasDOI"] = []  
                    pub["hasDOI"].append(identifier["id"])
                else : 
                    if identifier["type"] == "issn" :
                        if "hasISSN" not in pub:
                            pub["hasISSN"] = []                      
                        pub["hasISSN"].append(identifier["id"])
                    elif identifier["type"] == "isbn" :
                        if "hasISBN" not in pub:
                            pub["hasISBN"] = []                          
                        pub["hasISBN"].append(identifier["id"])
                
                if (("url" in identifier)) :
                    if "hasLink" not in pub:
                        pub["hasLink"] = []  
                    pub["hasLink"].append(identifier["url"])

            del pub["identifier"]
        
        return [pub, objhash, []]
    
    def _values_to_string(self, obj, objhash) :
        if "values" in obj :
            if (type(obj["values"]) is list) :
                obj["values"] = ", ".join(obj["values"])
        return [obj, objhash, []]

    
    def _guess_sensor_type(self, archive, observation, sensor) :
        if (('sensorGenus' in sensor) or ('sensorSpecies' in sensor)) :
            if (archive == "MarineSediment") :
                return "Foraminifera"
            elif (archive == "Coral") :
                return "Polyp"
            elif (archive == "Wood") :
                return "Vegetation"
            elif (archive == "MolluskShell") :
                return "Bivalves"
            elif (archive == "Sclerosponge") :
                return "Sponge"
            return "OrganicSensor"
        else : 
            if (archive == "MarineSediment" and (observation == "Uk37" or observation == "Alkenone")) :
                type = "Coccolithophores"
            elif (archive == "MarineSediment" and observation == "TEX86") :
                type = "Archea"
            elif (archive == "MarineSediment" and observation == "D18O") :
                type = "Foraminifera"
            elif (archive == "MarineSediment" and observation == "Mg/Ca") :
                type = "Foraminifera"
            elif (archive == "LakeSediment" and (observation == "Uk37" or observation == "Alkenone")) :
                type = "Coccolithophores"
            elif (archive == "LakeSediment" and observation == "TEX86") :
                type = "Archea"
            elif (archive == "LakeSediment" and observation == "Midge") :
                type = "Chironomids"
            elif (archive == "LakeSediment" and observation == "BSi") :
                type = "Diatoms"
            elif (archive == "LakeSediment" and observation == "Chironomid") :
                type = "Chironomids"
            elif (archive == "LakeSediment" and observation == "Reflectance") :
                type = "PhotosyntheticAlgae"
            elif (archive == "LakeSediment" and observation == "Pollen") :
                type = "Watershed"
            elif (archive == "Coral") :
                return "Polyp"
            elif (archive == "Wood") :
                return "Vegetation"
            elif (archive == "MolluskShell") :
                return "Bivalves"
            elif (archive == "Sclerosponge") :
                return "Sponge"
            elif (archive == "Speleothem") :
                return "Karst"
            elif (archive == "GlacierIce") :
                return "Snow"
            elif (archive == "LakeSediment" and observation == "VarveThickness") :
                return "Catchment"
            elif (archive == "GlacierIce" and observation == "Melt") :
                return "IceSurface"
            elif (archive == "Borehole") :
                return "Soil"
            else : 
                return "InorganicSensor"
    
    def _get_observation(self, observation) :
        if observation is None:
            return None
        if (observation.lower() == "alkenone") :
            return "Uk37"
        return camelCase(observation)

    def _get_variable_id(self, obj, parentid) :
        iobj = dict((k.lower(), v) for k, v in obj.items())
        if "tsid" not in iobj:
            iobj["tsid"] = uniqid()
        id =  parentid + "." + iobj["tsid"]
        id += "." + str(iobj.get("variablename", ""))
        return id
    
    def _wrap_integration_time(self, obj, objhash) :
        objid = obj["@id"]
        # Deal with integrationTime
        pvals = {}
        for key,value in obj.items() :
            if (re.search(r"^integrationTime\$", key, re.IGNORECASE)) :
                pvals["hasValue"] = value
                del obj[key]
            else:
                m = re.search(r"^integrationTime(.+)", key)
                if m is not None:
                    nkey = m.groups()[0]
                    nkey_lcfirst = lcfirst(nkey)
                    pvals[nkey_lcfirst] = value
                    del obj[key]

        if len(pvals.values()) > 0:
            intimeid = objid + '.IntegrationTime'
            obj['integrationTime'] = intimeid
            intime = {}
            intime['@id'] = intimeid
            intime['@category'] = 'IntegrationTime'
            intime['@schema'] = 'IntegrationTime'
            intime.update(pvals)
            objhash[intimeid] = intime
            return [obj, objhash, [intimeid]]
        
        return [obj, objhash, []]

    def _add_interpretation_rank(self, obj, objhash):
        if "rank" not in obj or type(obj["rank"]) != int:
            rank = obj["@index"] - 1
            obj["rank"] = rank
        return [obj, objhash, []]
    
    def _wrap_uncertainty(self, obj, objhash) :
        objid = obj["@id"]
        # Deal with uncertainty
        pvals = {}
        keys_to_be_deleted = []
        for key,value in obj.items() :
            if (re.search(r"^uncertainty\$", key, re.IGNORECASE)) :
                pvals["hasValue"] = value
                keys_to_be_deleted.append(key)
            elif (re.search(r"^uncertainty", key, re.IGNORECASE)) :
                pvals[key] = value
                keys_to_be_deleted.append(key)

        for key in keys_to_be_deleted:
            del obj[key]

        if len(pvals.values()) > 0 :
            uncid = "" + str(objid) + ".Uncertainty"
            obj["hasUncertainty"] = uncid
            uncertainty = {
                "@id": uncid,
                "@category": "Uncertainty"
            }
            for prop,value in pvals.items() :
                uncertainty[prop] = value
            
            objhash[uncid] = uncertainty
            return [obj, objhash, [uncid]]
        
        return [obj, objhash, []]
    
    def _add_found_in_table(self, obj, objhash) :
        obj["foundInTable"] = obj["@parent"]["@id"]
        return [obj, objhash, []]
    
    def _add_found_in_dataset(self, obj, objhash) :
        parent = obj["@parent"]
        top = parent
        while (parent) :
            top = parent
            parent = parent["@parent"]
        obj["foundInDataset"] = top["@id"]
        return [obj, objhash, []]
        
    # Unroll the list to a rdf first/rest structure
    def _unroll_values_list_to_rdf(self, lst: list, dtype):
        listitems = []
        for idx, item in enumerate(lst):
            listitems.append(Literal(item, datatype=(XSD[dtype] if dtype in XSD else None)))

        listid = BNode()
        list = Collection(self.graph, listid, listitems)
        return listid

    def _parse_changes(self, changes, parent=None):
        newChanges = []
        if not isinstance(changes, list):
            changes = [changes]
        
        for change in changes:
            for name in change.keys():
                notes = change[name] or []
                newChange = {
                    "name": name,
                    "notes": [item[0] for item in notes]
                }
                newChanges.append(newChange)
        return newChanges
    

    def _add_variable_values(self, obj, objhash) :
        if "filename" in obj["@parent"]:
            csvname = obj["@parent"]["filename"]
        else:
            csvname = re.sub(r"^.*\.", "", obj["@parent"]["@id"]) + ".csv"

        if "number" not in obj:
            obj["number"] = obj["@index"]
        if type(obj["number"]) is str:
            obj["number"] = int(obj["number"])
        if not isinstance(obj["number"], list):
            obj["number"] = [obj["number"]]
        
        indices = [int(col)-1 for col in obj["number"]]
        if csvname in self.lipd_csvs:
            df = self.lipd_csvs[csvname]
            values = []
            if len(indices) == 1:
                if indices[0] < len(df.columns):
                    df_values = df[indices[0]]
                    values = df_values.tolist()
                    #dtype = "float" if df_values.dtypes == "float64" else "string"
            else:
                df_values = df[indices]
                values = df_values.values.tolist()
                #dtype = "float" if df_values[0].dtypes == "float64" else "string"

            # TODO: Dumping to json string for now. 
            # - Compressing string if it is too long (otherwise loading to RDF hangs)
            valstring = json.dumps(values)
            #if len(valstring) > 1000000:
            #    valstring =json.dumps({"base64_zlib": zip_string(valstring)})
            obj["hasValues"] = valstring

            # rdf:Seq doesn't seem to be importing well in GraphDB            
            #bnodeid = self._unroll_values_list_to_rdf(values, dtype)
            #obj["hasValues"] = bnodeid

            return [obj, objhash, []]
        return [obj, objhash, []]

    def _add_standard_variable(self, obj, objhash) :
        if "variableName" in obj:
            name = obj["variableName"]
            synonyms = SYNONYMS["VARIABLES"]["PaleoVariable"]
            if type(name) is str and name.lower() in synonyms:
                obj["hasStandardVariable"] = synonyms[name.lower()]["id"]
                # Only add object label in the current graph if set
                if self.add_labels:
                    label = synonyms[name.lower()]["label"]                        
                    self._set_object_label(obj["hasStandardVariable"], label)

        return [obj, objhash, []]

    def _stringify_column_numbers_array(self, obj, objhash):
        if "number" in obj and isinstance(obj["number"], list) and len(obj["number"]) > 1:
            obj["hasColumnNumber"] = json.dumps(obj["number"])
            del obj["number"]
        return [obj, objhash, []]
    
    def _modify_structure_if_needed(self, obj, objhash, schema) :
        if (("@fromJson" in schema)) :
            for func in schema["@fromJson"]: 
                fn = getattr(self, func)
                (obj, objhash, newids) = fn(obj, objhash)
                for newid in newids : 
                    if ((newid in objhash)) :
                        newobj = objhash[newid]
                        if (type(newobj) is dict) and ("@category" in newobj) :
                            newschid = newobj["@category"]
                            newschema =  self.schema[newschid] if (newschid in self.schema) else {}
                            (objhash[newid], objhash) = self._modify_structure_if_needed(newobj, objhash, newschema)
        
        return [obj, objhash]
    
    def _get_compound_key_id(self, compound_key, obj) :
        tobj = obj
        for key in compound_key : 
            if ((type(tobj) is dict) and (key in tobj)) :
                tobj = tobj[key]
            else : 
                return None
            
        if not type(tobj) is dict:
            return tobj
        
        return None
    
    def _get_binding_key_id(self, key, obj) :
        key_options = key.split("|")
        for optkey in key_options : 
            compound_key = optkey.split(".")
            keyid = self._get_compound_key_id(compound_key, obj)
            if (keyid) :
                return keyid
        return uniqid()
    
    def _get_function_key_id(self, fn, arg, curobjid) :
        if (fn == "trunc") :
            return curobjid[0:0 + len(curobjid) - int(arg)]
        elif (fn == "uniqid") :
            return str(curobjid) + uniqid(arg)
        return curobjid
    
    def _create_id_from_pattern(self, pattern, obj) :
        objid = ""
        for key in pattern : 
            m = re.search(r"{(.+)}", key)
            if m and len(m.groups()) > 0 :
                objid += str(self._get_binding_key_id(m.groups()[0], obj))
            else : 
                m = re.search(r"_(.+)\((.*)\)", key)
                if m and len(m.groups()) > 1:
                    fn = m.groups()[0]
                    arg = m.groups()[1]
                    objid = str(self._get_function_key_id(fn, arg, objid))
                else : 
                    objid += str(key)
        return objid
    
    def _fix_title(self, titleid) :
        return titleid.replace(r"@\\x{FFFD}@u", '_')
    
    def _get_object_id(self, obj, category, schema) :
        if type(obj) is dict:
            objid =  "Unknown." + uniqid(category)
        else:
            objid = ucfirst(obj).replace(" ", "_")
        if (("@id" in schema)) :
            objid = self._create_id_from_pattern(schema["@id"], obj)
        
        return self._fix_title(objid)
    
    def _map_lipd_to_json(self, obj, parent, index, category, schemaname, hash) :
        schema =  self.schema[schemaname] if (schemaname in self.schema) else {}
        
        if not type(obj) is dict:
            return obj
        
        obj["@parent"] = parent
        obj["@index"] = index
        obj["@schema"] = schemaname
        
        objid = self._get_object_id(obj, category, schema)
        if (("@id" in obj)) :
            objid = obj["@id"]
        if ((objid in hash)) :
            return objid
        obj["@id"] = objid
        
        (obj, hash) = self._modify_structure_if_needed(obj, hash, schema)
        
        if ("@category" in obj) :
            category = obj["@category"]
        hash[objid] = {
            "@id": objid,
            "@category" : category,
            "@schema" : schemaname
        }
        item = hash[objid]
        
        if type(obj) is dict :
            for propkey,value in obj.items() :
                if (propkey[0] == "@") :
                    continue
                
                if propkey in BLACKLIST :
                    continue
                
                details = {}
                pname = propkey
                if propkey in schema :
                    details = schema[propkey]
                    pname =  details["name"] if ("name" in details) else propkey
                
                dtype =  details["type"] if ("type" in details) else None
                cat =  details["category"] if ("category" in details) else None
                sch =  details["schema"] if ("schema" in details) else None
                fromJson =  details["fromJson"] if ("fromJson" in details) else None
                if (sch and not cat) :
                    cat = sch
                if (fromJson) :
                    fn = getattr(self, fromJson)
                    value = fn(value, obj)
                    if (not value) :
                        continue
                    
                    if (pname) :
                        if (type(value) is list) :
                            index = 1
                            for subvalue in value:                              
                                if (type(subvalue) is dict):
                                    if propkey not in item:
                                        item[propkey] = []
                                    item[propkey].append(self._map_lipd_to_json(subvalue, obj, index, cat, sch, hash))
                                    index+=1
                        else : 
                            if (type(value) is dict):
                                item[propkey] = self._map_lipd_to_json(value, obj, None, cat, sch, hash)
                            else : 
                                item[propkey] = value
                            
                        
                    else : 
                        if (type(value) is dict):
                            for subpropkey,subvalue in value.items() :
                                item[subpropkey] = subvalue
                    continue
                
                if (not pname) :
                    continue
                
                if (type(value) is list):
                    index = 1
                    for subvalue in value: 
                        if propkey not in item:
                            item[propkey] = []                    
                        item[propkey].append(self._map_lipd_to_json(subvalue, obj, index, cat, sch, hash))
                        index+=1
                    
                else : 
                    if (type(value) is dict):
                        if propkey not in item:
                            item[propkey] = []                      
                        item[propkey].append(self._map_lipd_to_json(value, obj, None, cat, sch, hash))
                    else : 
                        if (dtype == "Individual") :
                            item[propkey] = value
                            if (not (value in hash)) :
                                hash[value] = {
                                    "@id" : value,
                                    "@category" : cat,
                                    "@schema" : sch
                                }
                        else : 
                            item[propkey] = value

        hash[objid] = item
        return objid
    
    def _guess_data_value_type(self, val) :
        value = str(val)
        if (re.search(r"^-?\d+$", value)) :
            return "float" #"integer"
        
        if (re.search(r"^-?\d+\.\d+$", value)) :
            return "float"
        
        if (re.search(r"^[2][0-9]{3}[-][0-1][0-9][-][0-3][0-9]( |T)[0-9]{2}:[0-9]{2}:[0-9]{2}", value)) :
            return "datetime"
                
        if (re.search(r"^[2][0-9]{3}[-][0-1][0-9][-][0-3][0-9]", value)) :
            return "date"
        
        if (re.search(r"^(true|false)$", value, re.IGNORECASE)) :
            return "boolean"
        
        if (re.search(r"^http", value)) :
            return "url"
        
        #if (re.search(r"^.+@.+\..+", value)) :
        #    return "Email"
        
        if (re.search(r"^\".+\"$", value)) :
            return "string"
        
        if (re.search(r"^'.+'$", value)) :
            return "string"
        
        return "string"
    
    def _guessValueType(self, value) :
        if value:
            if type(value) is list :
                for subvalue in value :
                    return self._guessValueType(subvalue)
            elif type(value) is dict : 
                return "Individual"
            
            else : 
                valtype = self._guess_data_value_type(value)
                return valtype

        return "string"
    
    def _get_property_details(self, key, schema, value) :
        pname = key
        details = {
            "name": pname
        }
        if (key in schema) and ("@@processed" in schema[key]) :
            return schema[key]
        
        # Get details from schema
        if (key in schema) :
            for skey,svalue in schema[key].items() :
                details[skey] = svalue
        
        if (("schema" in details)) :
            details["type"] = "Individual"
        
        pname = lcfirst(details["name"])
        
        # Get more details from the property definition (if it exists)
        """
        newname = resolveProperty(pname)
        if (newname) :
            details["type"] = getOntPropertyRange(newname)
            details["name"] = newname
        """ 
        
        if (not ("type" in details)) :
            details["type"] = self._guessValueType(value)
            if (not ("type" in details)) :
                details["type"] = "string"

        details["@@processed"] = True
        schema[key] = details
        return details
    
    # Create individual
    def _create_individual(self, objid) :
        return self.namespace + sanitizeId(objid)
    
    # Create class
    def _create_class(self, category) :
        return ONTONS + sanitizeId(category)
    
    # Create property
    def _create_property(self, prop, dtype, cat, multiple) :
        nsprop = prop.split(":", 2)
        ns = ONTONS
        if len(nsprop) > 1 :
            prefix = nsprop[0]
            if prefix in NAMESPACES:
                ns = NAMESPACES[prefix]
            prop = nsprop[1]
        return [ ns + lcfirst(sanitizeId(prop)), dtype, cat, multiple ]
    
    # Set individual classes
    def _set_individual_classes(self, objid, category, extracats) :
        if objid and category:
            self.graph.add((
                URIRef(objid),
                RDF.type,
                URIRef(category),
                URIRef(self.graphurl)
            ))
        for ecat in extracats:
            if objid and ecat:
                self.graph.add((
                    URIRef(objid),
                    RDF.type,
                    URIRef(self._create_class(ecat)),
                    URIRef(self.graphurl)
                ))
    
    # Set object label
    def _set_object_label(self, objid, label) :
        if objid and label:
            self.graph.add((
                URIRef(objid),
                RDFS.label,
                Literal(label),
                URIRef(self.graphurl)
            ))

    # Set property value
    def _set_property_value( self, objid, prop, value ):
        if (type(value) is list) :
            for subvalue in value : 
                self._set_property_value(objid, prop, subvalue)
            return

        (propid, dtype, cat, multiple) = prop
        if objid and value:
            objitem = None

            if dtype == "float" or dtype == "integer":
                if re.search("^.*[^a-zA-Z]?nan[^a-zA-Z]?.*$", str(value).lower()):
                    return
                if re.search("^.*[^a-zA-Z]?na[^a-zA-Z]?.*$", str(value).lower()):
                    return

            if type(value) is str:
                value = escape(value)

            if dtype == "boolean":
                value = str(value).lower()
                if value != "true":
                    value = "false"
            
            elif dtype == "float":
                m = re.search(r"(\-?\d+\.?\d*)", str(value))
                if m:
                    value = m.group(1)
                else:
                    value = 0.0

            elif dtype == "integer":
                m = re.search(r"(\-?\d+)", str(value))
                if m:
                    value = m.group(1)
                else:
                    value = 0

            if dtype == "Individual":
                value = self._create_individual(value)
                objitem = URIRef(value)
            
            elif dtype == "EnumeratedIndividual":
                objitem = URIRef(value)
            
            elif dtype == "List":
                objitem = value
            
            else:
                objitem = Literal(value, datatype=(XSD[dtype] if dtype in XSD else None))

            # FIXME: Do not add if property doesn't allow multiple values and a value already exists
            if not multiple:
                existing = list(self.graph.triples((URIRef(objid), URIRef(propid), None)))
                if len(existing) > 0:
                    return
            
            self.graph.add((
                URIRef(objid),
                URIRef(propid),
                objitem,
                URIRef(self.graphurl)
            ))
    
    def _create_individual_full(self, obj) :
        category = obj["@category"]
        extracats =  obj["@extracats"] if ("@extracats" in obj) else {}
        schemaname =  obj["@schema"] if ("@schema" in obj) else category
        schema =  self.schema[schemaname] if (schemaname in self.schema) else {}
        objid = obj["@id"]
        if (not objid) :
            return
        
        # Create category
        if (category) :
            category = self._create_class(category)
        
        objid = self._create_individual(objid)
        
        # Set Individual classes
        self._set_individual_classes(objid, category, extracats)
        
        for key,value in obj.items() :
            if (key[0] == "@") :
                continue
            
            details = self._get_property_details(key, schema, value)
            prop = details["name"]
            dtype = details["type"]
            synonyms = details.get("synonyms", {})
            cat =  details["category"] if ("category" in details) else None
            sch =  details["schema"] if ("schema" in details) else None
            if (sch and not cat) :
                cat = sch
            
            fromJson =  details["fromJson"] if ("fromJson" in details) else None
            multiple =  details["multiple"] if ("multiple" in details) else False
            if (not prop) :
                continue
                
            # Create Property
            propDI = self._create_property(prop, dtype, cat, multiple)

            # Set property value
            if dtype == "Individual":
                if type(value) is str and synonyms:
                    # If the value is a string and there are synonyms for this Individual
                    if value.lower() in synonyms:
                        # If we have a synonym-mapping for the value to an Individual
                        
                        propDI[1] = "EnumeratedIndividual" # Rename property type to be an enumeration
                        synid = synonyms[value.lower()]["id"]
                        if not self.standardize:
                            # If we don't want to standardize, then create a unique id for the individual
                            synid += "." + uniqid()
                        self._set_property_value(objid, propDI, synid)
                        # Only add object label in the current graph if set
                        if self.add_labels:
                            if self.standardize:
                                # Set the standard label for the individual
                                label = synonyms[value.lower()]["label"]
                            else:
                                # Set the user label for the individual
                                label = value
                            self._set_object_label(synid, label)
                    else:
                        # We don't have a synonym-mapping for the value. Create an individual and set its label to the value
                        propDI[1] = "EnumeratedIndividual"
                        synid = self._create_individual(value) + "." + uniqid()
                        self._set_property_value(objid, propDI, synid)
                        self._set_object_label(synid, value)
                else:
                    # There are no synonyms, and value is not a string. Just use it directly
                    self._set_property_value(objid, propDI, value)
            elif type(value) is dict:
                self._set_property_value(objid, propDI, value)
            else : 
                if (dtype == "File") :
                    # Enable this ?
                    """
                    fileid = uploadFile(value)
                    if (fileid) :
                        protectIndividual(fileid)
                        data = set_property(data, propDI, fileid)
                    """

                else : 
                    self._set_property_value(objid, propDI, value)

    def _find_files_with_extension(self, directory, extension):
        myregexobj = re.compile(r'\.'+extension+'$')
        try: 
            for entry in os.scandir(directory):
                if entry.is_file() and myregexobj.search(entry.path): 
                    yield entry.path, entry.name
                elif entry.is_dir():   # if its a directory, then repeat process as a nested function
                    yield from self._find_files_with_extension(entry.path, extension)
        except OSError as ose:
            print('Cannot access ' + directory +'. Probably a permissions error ', ose)

    def _load_lipd_json_to_graph(self, jsonpath, url=None):
        self.graph = ConjunctiveGraph()
        objhash = {}
        
        with open(jsonpath) as f:
            obj = json.load(f)
            if "dataSetName" in obj:
                self.graphurl = NSURL + "/" + sanitizeId(obj["dataSetName"])
            
            self._map_lipd_to_json(obj, None, None, "Dataset", "Dataset", objhash)
            if url:
                objhash[obj["@id"]]["hasUrl"] = url
            else:
                objhash[obj["@id"]]["hasUrl"] = DATAURL + "/" + obj["@id"] + ".lpd"

            for key, item in objhash.items():
                self._create_individual_full(item)


from tqdm import tqdm
from .globals.queries import QUERY_FILTER_VARIABLE_NAME, QUERY_VARIABLE, QUERY_DISTINCT_VARIABLE, QUERY_VARIABLE_ESSENTIALS, QUERY_DISTINCT_PROXY, QUERY_FILTER_VARIABLE_PROXY, QUERY_FILTER_VARIABLE_RESOLUTION, QUERY_LiPDSERIES_PROPERTIES
from .utils.multi_processing import multi_load_lipd_series
from .utils.rdf_graph import RDFGraph

import numpy as np
import json

class LiPDSeries(RDFGraph):
    '''The LiPD Series class describes a collection of `LiPD (Linked Paleo Data) <https://cp.copernicus.org/articles/12/1093/2016/cp-12-1093-2016.html>`_ 
    variables. It contains an `RDF <https://www.w3.org/RDF/>`_ Graph which is serialization of  LiPD variables into an RDF graph containing terms from 
    the `LiPD Ontology <http://linked.earth/Ontology/release/core/1.2.0/index-en.html>`_. Each LiPD Variable is also associated with the LiPD itself
    so it can be deserialized into the original LiPD format.
    How to browse and query the LiPD variables is described in a short example below.

    Examples
    --------
    In this example, we read an online LiPD file and convert it into a time series object dictionary.

    .. jupyter-execute::

        from pylipd.lipd_series import LiPDSeries

        lipd = LiPD()
        lipd.load(["https://lipdverse.org/data/LCf20b99dfe8d78840ca60dfb1f832b9ec/1_0_1//Nunalleq.Ledger.2018.lpd"])
        lipd_series = lipd.to_lipd_series()
    '''
    def __init__(self, graph=None):
        super().__init__(graph)
        self.lipds = {}


    def load(self, lipd, parallel=False):
        '''Extract Variables from the LiPD object.

        Parameters
        ----------
        lipd : LiPD
            A LiPD object
        
               
        Examples
        --------
        .. jupyter-execute::

            from pylipd.lipd_series import LiPDSeries

            lipd = LiPD()
            lipd.load(["https://lipdverse.org/data/LCf20b99dfe8d78840ca60dfb1f832b9ec/1_0_1//Nunalleq.Ledger.2018.lpd"])
            lipd_series = lipd.to_lipd_series()
        '''

        print(f"Creating LiPD Series...")

        # Update graph (Create contexts for each variable)
        print("- Extracting dataset subgraphs")
        total = len(lipd.get_all_dataset_names())
        for ctx in tqdm(lipd.graph.contexts(), total=total):
            ctxid = str(ctx.identifier)
            self.lipds[ctxid] = lipd.get(ctxid)
                    
        multi_load_lipd_series(self.graph, self.lipds, parallel)
        
        print("Done..")


    def get_all_variables(self):
        '''
        Returns a list of all variables in the graph
        
        Returns
        -------

        pandas.DataFrame
            A dataframe of all variables in the graph with columns uri, varid, varname
        
        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.utils.dataset import load_dir

            lipd = load_dir()
            S = lipd.to_lipd_series()
            df = S.get_all_variables()
            
            print(df)
        
        
        '''        
        return self.query(QUERY_VARIABLE)[1]
    
    def get_all_variable_names(self):
        
        """
        Get a list of all possible distinct variableNames. Useful for filtering and querying. 

        Returns
        -------
        list
            A list of unique variableName 
        
        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir
            lipd = load_dir('Pages2k')
            S = lipd.to_lipd_series()
            varName = S.get_all_variable_names()
            print(varName)
        """

        return self.query(QUERY_DISTINCT_VARIABLE)[1].iloc[:,0].values.tolist()
    
    def get_all_proxy(self):
        
        """
        Get a list of all possible proxy. Useful for filtering and querying. 

        Returns
        -------
        list
            A list of unique proxies
        
        Examples
        --------
        
        .. jupyter-execute::
            
            from pylipd.utils.dataset import load_dir
            lipd = load_dir('Pages2k')
            S = lipd.to_lipd_series()
            proxyName = S.get_all_proxy()
            print(proxyName)
        """
        
        return self.query(QUERY_DISTINCT_PROXY)[1].iloc[:,0].values.tolist()

    def get_timeseries_essentials(self):
        '''This function returns information about each variable: `dataSetName`, `archiveType`, `name`, `values`, `units`, `TSID`, `proxy`.

        Returns
        -------
        qres_df : pandas.DataFrame
            A dataframe containing the information in each column
        
        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.utils.dataset import load_dir

            lipd = load_dir()
            S = lipd.to_lipd_series()
            df = S.get_timeseries_essentials()
            
            print(df)

        '''
        
    
        query = QUERY_VARIABLE_ESSENTIALS
        qres, qres_df = self.query(query)
        
        #fix the dataframe
        for _,row in qres_df.iterrows():
            string = row['dataSetName'].split('/')[-1]
            row['dataSetName'] = string
        
        qres_df['values']=qres_df['values'].apply(lambda row : np.array(json.loads(row)))

        return qres_df
    
    def get_variable_properties(self):
        """
        Get a list of all the properties name associated with the dataset. Useful to write custom queries

        Returns
        -------
        clean_list : list
            A list of unique variable properties
        
        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.utils.dataset import load_dir

            lipd = load_dir()
            S = lipd.to_lipd_series()
            l = S.get_variable_properties()
            
            print(l)


        """
        
        query_list = self.query(QUERY_LiPDSERIES_PROPERTIES)[1].iloc[:,0].values.tolist()
        clean_list = [item.split("#")[-1] for item in query_list]
        
        return clean_list
    

    def filter_by_name(self, name):
        '''
        Filters series to return a new LiPDSeries that only keeps variables that have the specified name (regex)

        Parameters
        ----------

        name : str
            The variable name to filter by

        Returns
        -------
        
        pylipd.lipd_series.LiPDSeries
            A new LiPDSeries object that only contains variables that have the specified name (regex)
        
        Examples
        --------
        
        .. jupyter-execute::

            from pylipd.utils.dataset import load_datasets
            lipd = load_datasets('ODP846.Lawrence.2006.lpd')
            S = lipd.to_lipd_series()
            sst = S.filter_by_name('sst')
            
            print(sst.get_all_variable_names())

        '''
        query = QUERY_FILTER_VARIABLE_NAME
        query = query.replace("[name]", name)

        qres, qres_df = self.query(query)
        varuris = [str(row.uri) for row in qres]
        dsuris = [*set([str(row.dsuri) for row in qres])]

        #print(len(dsuris))

        rdfgraph = self.get(varuris)
        S = LiPDSeries(rdfgraph.graph)
        S.lipds = {k: self.lipds[k].copy() for k in dsuris}
        return S

    def filter_by_proxy(self, proxy):
        '''
        Filters series to return a new LiPDSeries that only keeps variables that have the specified proxy (regex)
    
        Parameters
        ----------
    
        proxy : str
            The name of the proxy to filter by
    
        Returns
        -------
        
        pylipd.lipd_series.LiPDSeries
            A new LiPDSeries object that only contains variables that have the specified name (regex)
        
        Examples
        --------
        
        .. jupyter-execute::
    
            from pylipd.utils.dataset import load_dir
            lipd = load_dir('Pages2k')
            S = lipd.to_lipd_series()
            S_filtered = S.filter_by_proxy('ring width')
            print(S_filtered.get_all_proxy())
    
        '''
        query = QUERY_FILTER_VARIABLE_PROXY
        query = query.replace("[proxy]", proxy)
    
        qres, qres_df = self.query(query)
        varuris = [str(row.uri) for row in qres]
        dsuris = [*set([str(row.dsuri) for row in qres])]
    
        rdfgraph = self.get(varuris)
        S = LiPDSeries(rdfgraph.graph)
        S.lipds = {k: self.lipds[k].copy() for k in dsuris}
        return S
    
    def filter_by_resolution(self, threshold, stats='Mean'):
        '''
        Filters series to return a new LiPDSeries that only keeps variables that have a resolution less than the specified threshold. 

        Parameters
        ----------
        threshold : float
            The maximum resolution to keep
        stats : str, optional
            Whether to use 'Mean', 'Median', 'Min' or 'Max' resolution. The default is 'Mean'.

        Raises
        ------
        ValueError
            Make sure that the stats is of ['Mean','Median', 'Min', 'Max'].

        Returns
        -------
        S : pylipd.lipd_series.LiPDSeries
            A new LiPDSeries object that only contains the filtered variables
        
        Examples
        --------
        
        .. jupyter-execute::
    
            from pylipd.utils.dataset import load_dir
            lipd = load_dir('Pages2k')
            S = lipd.to_lipd_series()
            S_filtered = S.filter_by_resolution(10)

        '''
        
        stats = stats.capitalize() #make sure that the first letter is capitalized
        stats_allowed = ['Mean','Median', 'Min', 'Max'] #possible values
        if stats not in stats_allowed:
            raise ValueError("Stats must be ['Mean','Median', 'Min', 'Max']")
        
        threshold = float(threshold) # make sure this is a float or can be coerced in one
        
        query = QUERY_FILTER_VARIABLE_RESOLUTION
        query = query.replace("[value]", str(threshold))
        query = query.replace("[stat]", stats)

        qres,q_df = self.query(query)
        
        varuris = [str(row.uri) for row in qres]
        dsuris = [*set([str(row.dsuri) for row in qres])]
    
        rdfgraph = self.get(varuris)
        S = LiPDSeries(rdfgraph.graph)
        S.lipds = {k: self.lipds[k].copy() for k in dsuris}
        return S



##############################
# Auto-generated. Do not Edit
##############################
from pylipd.globals.synonyms import SYNONYMS

class PaleoProxyGeneral:
    """Controlled-vocabulary class for `PaleoProxyGeneral` terms."""
    synonyms = SYNONYMS["PROXIES"]["PaleoProxyGeneral"]

    def __init__(self, id, label):
        """Initialize a PaleoProxyGeneral term.

        Parameters
        ----------
        id : str
            The full URI identifier for this term.
        label : str
            The human-readable label for this term.
        """
        self.id = id
        self.label = label
    
    def __eq__(self, value: object) -> bool:
        """Check equality based on term ID.

        Parameters
        ----------
        value : object
            The object to compare against.

        Returns
        -------
        bool
            True if the IDs match, False otherwise.
        """
        return self.id == value.id
        
    def getLabel(self):
        """Return the human-readable label of the term.

        Returns
        -------
        str
            The label for this term.
        """
        return self.label

    def getId(self):
        """Return the full URI identifier of the term.

        Returns
        -------
        str
            The URI identifier for this term.
        """
        return self.id
    
    def to_data(self, data={}):
        """Serialize this term to JSON-LD format.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] ={
            "label": [
                {
                    "@datatype": None,
                    "@type": "literal",
                    "@value": self.label
                }
            ]
        }
        return data

    def to_json(self):
        """Return the plain-text label (used in lightweight JSON).

        Returns
        -------
        str
            The label for this term.
        """
        data = self.label
        return data

    @classmethod
    def from_synonym(cls, synonym):
        """Create a PaleoProxyGeneral instance from a synonym string.

        Parameters
        ----------
        synonym : str
            A synonym or alternative name for the term.

        Returns
        -------
        PaleoProxyGeneral or None
            The PaleoProxyGeneral instance if found, None otherwise.
        """
        if synonym.lower() in PaleoProxyGeneral.synonyms:
            synobj = PaleoProxyGeneral.synonyms[synonym.lower()]
            return PaleoProxyGeneral(synobj['id'], synobj['label'])
        return None
        
class PaleoProxyGeneralConstants:
    biogenic = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#biogenic", "biogenic")
    cryophysical = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#cryophysical", "cryophysical")
    dendrophysical = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#dendrophysical", "dendrophysical")
    elemental = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#elemental", "elemental")
    faunal_assemblage = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#faunal_assemblage", "faunal assemblage")
    floral_assemblage = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#floral_assemblage", "floral assemblage")
    isotopic = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#isotopic", "isotopic")
    mineral = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#mineral", "mineral")
    pyrogenic = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#pyrogenic", "pyrogenic")
    sedimentology = PaleoProxyGeneral("http://linked.earth/ontology/paleo_proxy#sedimentology", "sedimentology")


"""
Legacy code from the LiPD utils in Python
"""

import copy
import re

from ..series.regexes import re_pandas_x_und, re_sheet
EMPTY = ['', ' ', None, 'na', 'n/a', '?', "'", "''"]

class LiPD_Legacy:
    ##############################################
    # TODO: Create LiPDSeries and MultipleLiPDSeries objects
    # LiPDSeries is:
    # - Age and/or Year and/or Depth (values) & a Variable (values) + Dataset id + Table id + Variable id 
    # - Look at : https://github.com/LinkedEarth/Pyleoclim_util/blob/master/pyleoclim/core/lipdseries.py
    # - use xAxisTs function (to return depth as well)
    #
    # MultipleLiPDSeries is:
    # - A list of LiPDSeries (could be from one or more datasets)
    ##############################################

    def extract(self, d, whichtables="meas", mode="paleo", time="age"):
        """
        LiPD Version 1.3
        Main function to initiate LiPD to TSOs conversion.

        Each object has a
        "paleoNumber" or "chronNumber"
        "tableNumber"
        "modelNumber"
        "time_id"
        "mode" - chronData or paleoData
        "tableType" - "meas" "ens" "summ"

        :param dict d: Metadata for one LiPD file
        :param str whichtables: all, meas, summ, or ens
        :param str mode: paleo or chron mode
        :return list _ts: Time series
        """
        _root = {}
        _ts = {}
        # _switch = {"paleoData": "chronData", "chronData": "paleoData"}
        _pc = "paleoData"
        if mode == "chron":
            _pc = "chronData"
        _root["mode"] = _pc
        _root["time_id"] = time
        if "archiveType" in d:
            _root["archiveType"] = d["archiveType"]
        try:
            # Build the root level data.
            # This will serve as the template for which column data will be added onto later.
            for k, v in d.items():
                if k == "funding":
                    _root = self._extract_fund(v, _root)
                elif k == "geo":
                    _root = self._extract_geo(v, _root)
                elif k == 'pub':
                    _root = self._extract_pub(v, _root)
                # elif k in ["chronData", "paleoData"]:
                #     # Store chronData and paleoData as-is. Need it to collapse without data loss.
                #     _root[k] = copy.deepcopy(v)
                else:
                    if k not in ["chronData", "paleoData"]:
                        _root[k] = v
            # Create tso dictionaries for each individual column (build on root data)
            _ts = self._extract_pc(d, _root, _pc, whichtables)
        except Exception as e:
            raise(e)

        return _ts


    def _extract_fund(self, l, _root):
        """
        Creates flat funding dictionary.
        :param list l: Funding entries
        """
        for idx, i in enumerate(l):
            for k, v in i.items():
                _root['funding' + str(idx + 1) + '_' + k] = v
        return _root


    def _extract_geo(self, d, _root):
        """
        Extract geo data from input
        :param dict d: Geo
        :return dict _root: Root data
        """
        # May not need these if the key names are corrected in the future.
        # COORDINATE ORDER: [LON, LAT, ELEV]
        x = ['geo_meanLon', 'geo_meanLat', 'geo_meanElev']
        # Iterate through geo dictionary
        for k, v in d.items():
            # Case 1: Coordinates special naming
            if k == 'coordinates':
                for idx, p in enumerate(v):
                    try:
                        # Check that our value is not in EMPTY.
                        if isinstance(p, str):
                            if p.lower() in EMPTY:
                                # If elevation is a string or 0, don't record it
                                if idx != 2:
                                    # If long or lat is empty, set it as 0 instead
                                    _root[x[idx]] = 0
                            else:
                                # Set the value as a float into its entry.
                                _root[x[idx]] = float(p)
                        # Value is a normal number, or string representation of a number
                        else:
                            # Set the value as a float into its entry.
                            _root[x[idx]] = float(p)
                    except IndexError as e:
                        raise e
            # Case 2: Any value that is a string can be added as-is
            elif isinstance(v, str):
                if k == 'meanElev':
                    try:
                        # Some data sets have meanElev listed under properties for some reason.
                        _root['geo_' + k] = float(v)
                    except ValueError as e:
                        # If the value is a string, then we don't want it
                        raise e
                else:
                    _root['geo_' + k] = v
            # Case 3: Nested dictionary. Recursion
            elif isinstance(v, dict):
                _root = self._extract_geo(v, _root)
        return _root


    def _extract_pub(self, l, _root):
        """
        Extract publication data from one or more publication entries.
        :param list l: Publication
        :return dict _root: Root data
        """
        # For each publication entry
        for idx, pub in enumerate(l):
            # Get author data first, since that's the most ambiguously structured data.
            _root = self._extract_authors(pub, idx, _root)
            # Go through data of this publication
            for k, v in pub.items():
                # Case 1: DOI ID. Don't need the rest of 'identifier' dict
                if k == 'identifier':
                    try:
                        _root['pub' + str(idx + 1) + '_DOI'] = v
                    except KeyError as e:
                        raise e
                # Case 2: All other string entries
                else:
                    if k != 'authors' and k != 'author':
                        _root['pub' + str(idx + 1) + '_' + k] = v
        return _root


    def _extract_authors(self, pub, idx, _root):
        """
        Create a concatenated string of author names. Separate names with semi-colons.
        :param any pub: Publication author structure is ambiguous
        :param int idx: Index number of Pub
        """
        try:
            # DOI Author data. We'd prefer to have this first.
            names = pub['author']
        except KeyError as e:
            try:
                # Manually entered author data. This is second best.
                names = pub['authors']
            except KeyError as e:
                # Couldn't find any author data. Skip it altogether.
                names = False

        # If there is author data, find out what type it is
        if names:
            # Build author names onto empty string
            auth = ''
            # Is it a list of dicts or a list of strings? Could be either
            # Authors: Stored as a list of dictionaries or list of strings
            if isinstance(names, list):
                for name in names:
                    if isinstance(name, str):
                        auth += name + ';'
                    elif isinstance(name, dict):
                        for k, v in name.items():
                            auth += str(v) + ';'
            elif isinstance(names, str):
                auth = names
            # Enter finished author string into target
            _root['pub' + str(idx + 1) + '_author'] = auth[:-1]
        return _root


    def _extract_pc(self, d, root, pc, whichtables):
        """
        Extract all data from a PaleoData dictionary.
        :param dict d: PaleoData dictionary
        :param dict root: Time series root data
        :param str pc: paleoData or chronData
        :param str whichtables: all, meas, summ, or ens
        :return list _ts: Time series
        """
        _ts = []
        try:
            # For each table in pc
            for v in d[pc]:
                if whichtables == "all" or whichtables == "meas":
                    for _table_data1 in v["measurementTable"]:
                        _ts = self._extract_table(_table_data1, copy.deepcopy(root), pc, _ts, "meas")
                if whichtables != "meas":
                    if "model" in v:
                        for _table_data1 in v["model"]:
                            # get the method info for this model. This will be paired to all summ and ens table data
                            _method = self._extract_method(_table_data1["method"])
                            if whichtables == "all" or whichtables == "summ":
                                if "summaryTable" in _table_data1:
                                    for _table_data2 in _table_data1["summaryTable"]:
                                        # take a copy of this tso root
                                        _tso = copy.deepcopy(root)
                                        # add in the method details
                                        _tso.update(_method)
                                        # add in the table details
                                        _ts = self._extract_table(_table_data2, _tso, pc, _ts, "summ")
                            if whichtables == "all" or whichtables == "ens":
                                if "ensembleTable" in _table_data1:
                                    for _table_data2 in _table_data1["ensembleTable"]:
                                        _tso = copy.deepcopy(root)
                                        _tso.update(_method)
                                        _ts = self._extract_table(_table_data2, _tso, pc, _ts, "ens")

        except Exception as e:
            raise e
        return _ts


    def _extract_method(self, method):
        """
        Make a timeseries-formatted version of model method data

        :param dict method: Method data
        :return dict _method: Method data, formatted
        """
        _method = {}
        for k,v in method.items():
            _method["method_" + k] = v
        return _method


    def _extract_special(self, current, table_data):
        """
        Extract year, age, and depth column from table data
        :param dict table_data: Data at the table level
        :param dict current: Current data
        :return dict current:
        """
        try:
            # Add age, year, and depth columns to ts_root where possible
            for v in table_data['columns']:
                if "variableName" not in v:
                    continue
                
                s = ""
                k = v["variableName"]

                # special case for year bp, or any variation of it. Translate key to "age""
                if type(k) is list:
                    k = k[0]
                    
                if "bp" in k.lower():
                    s = "age"

                # all other normal cases. clean key and set key.
                elif any(x in k.lower() for x in ('age', 'depth', 'year', "yr", "distance_from_top", "distance")):
                    # Some keys have units hanging on them (i.e. 'year_ad', 'depth_cm'). We don't want units on the keys
                    if re_pandas_x_und.match(k):
                        s = k.split('_')[0]
                    elif "distance" in k:
                        s = "depth"
                    else:
                        s = k

                # create the entry in ts_root.
                if s:
                    try:
                        if 'values' in v:
                            current[s] = v['values']
                    except KeyError as e:
                        # Values key was not found.
                        raise e
                    try:
                        if 'units' in v:
                            current[s + 'Units'] = v['units']
                    except KeyError as e:
                        # Values key was not found.
                        raise e

        except Exception as e:
            raise e

        return current


    def _extract_table_root(self, d, current, pc):
        """
        Extract data from the root level of a paleoData table.
        :param dict d: paleoData table
        :param dict current: Current root data
        :param str pc: paleoData or chronData
        :return dict current: Current root data
        """
        try:
            for k, v in d.items():
                if isinstance(v, str):
                    current[pc + '_' + k] = v
        except Exception as e:
            raise e
        return current


    def _extract_table_model(self, table_data, current, tt):
        """
        Add in modelNumber and summaryNumber fields if this is a summary table

        :param dict table_data: Table data
        :param dict current: LiPD root data
        :param str tt: Table type "summ", "ens", "meas"
        :return dict current: Current root data
        """
        try:
            if tt in ["summ", "ens"]:
                m = re.match(re_sheet, table_data["tableName"])
                if m:
                    _pc_num= m.group(1) + "Number"
                    current[_pc_num] = m.group(2)
                    current["modelNumber"] = m.group(4)
                    current["tableNumber"] = m.group(6)
                else:
                    print("extract_table_summary: Unable to parse paleo/model/table numbers")
        except Exception as e:
            print("extract_table_summary: {}".format(e))
        return current


    def _extract_table(self, table_data, current, pc, ts, tt):
        """
        Use the given table data to create a time series entry for each column in the table.

        :param dict table_data: Table data
        :param dict current: LiPD root data
        :param str pc: paleoData or chronData
        :param list ts: Time series (so far)
        :param bool summary: Summary Table or not
        :return list ts: Time series (so far)
        """
        current["tableType"] = tt
        # Get root items for this table
        current = self._extract_table_root(table_data, current, pc)
        # Add in modelNumber and tableNumber if this is "ens" or "summ" table
        current = self._extract_table_model(table_data, current, tt)
        # Add age, depth, and year columns to root if available
        _table_tmp = self._extract_special(current, table_data)
        try:
            # Start creating entries using dictionary copies.
            for _col_data in table_data["columns"]:
                # Add column data onto root items. Copy so we don't ruin original data
                _col_tmp = self._extract_columns(_col_data, copy.deepcopy(_table_tmp), pc)
                try:
                    ts.append(_col_tmp)
                except Exception as e:
                    print("extract_table: Unable to create ts entry, {}".format(e))
        except Exception as e:
            raise e
        return ts


    def _extract_columns(self, d, tmp_tso, pc):
        """
        Extract data from one paleoData column
        :param dict d: Column dictionary
        :param dict tmp_tso: TSO dictionary with only root items
        :return dict: Finished TSO
        """
        for k, v in d.items():
            if isinstance(v, dict):
                flat_data = self._extract_nested(pc + "_" + k, v, {})
                for n,m in flat_data.items():
                    tmp_tso[n] = m
            else:
                # Assume if it's not a special nested case, then it's a string value
                tmp_tso[pc + '_' + k] = v
        return tmp_tso


    def _extract_nested(self, crumbs, dat, flat_dat):
        try:
            for k, v in dat.items():
                if isinstance(v, dict):
                    flat_dat = self._extract_nested(crumbs + "_" + k, v, flat_dat)
                else:
                    flat_dat[crumbs + "_" + k] = v
        except Exception as e:
            print("ts: _extract_nested: " + e)

        return flat_dat



##############################
# Auto-generated. Do not Edit
##############################
from pylipd.globals.synonyms import SYNONYMS

class ArchiveType:
    """Controlled-vocabulary class for `ArchiveType` terms."""
    synonyms = SYNONYMS["ARCHIVES"]["ArchiveType"]

    def __init__(self, id, label):
        """Initialize a ArchiveType term.

        Parameters
        ----------
        id : str
            The full URI identifier for this term.
        label : str
            The human-readable label for this term.
        """
        self.id = id
        self.label = label
    
    def __eq__(self, value: object) -> bool:
        """Check equality based on term ID.

        Parameters
        ----------
        value : object
            The object to compare against.

        Returns
        -------
        bool
            True if the IDs match, False otherwise.
        """
        return self.id == value.id
        
    def getLabel(self):
        """Return the human-readable label of the term.

        Returns
        -------
        str
            The label for this term.
        """
        return self.label

    def getId(self):
        """Return the full URI identifier of the term.

        Returns
        -------
        str
            The URI identifier for this term.
        """
        return self.id
    
    def to_data(self, data={}):
        """Serialize this term to JSON-LD format.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] ={
            "label": [
                {
                    "@datatype": None,
                    "@type": "literal",
                    "@value": self.label
                }
            ]
        }
        return data

    def to_json(self):
        """Return the plain-text label (used in lightweight JSON).

        Returns
        -------
        str
            The label for this term.
        """
        data = self.label
        return data

    @classmethod
    def from_synonym(cls, synonym):
        """Create a ArchiveType instance from a synonym string.

        Parameters
        ----------
        synonym : str
            A synonym or alternative name for the term.

        Returns
        -------
        ArchiveType or None
            The ArchiveType instance if found, None otherwise.
        """
        if synonym.lower() in ArchiveType.synonyms:
            synobj = ArchiveType.synonyms[synonym.lower()]
            return ArchiveType(synobj['id'], synobj['label'])
        return None
        
class ArchiveTypeConstants:
    Borehole = ArchiveType("http://linked.earth/ontology/archive#Borehole", "Borehole")
    Coral = ArchiveType("http://linked.earth/ontology/archive#Coral", "Coral")
    FluvialSediment = ArchiveType("http://linked.earth/ontology/archive#FluvialSediment", "Fluvial sediment")
    GlacierIce = ArchiveType("http://linked.earth/ontology/archive#GlacierIce", "Glacier ice")
    GroundIce = ArchiveType("http://linked.earth/ontology/archive#GroundIce", "Ground ice")
    LakeSediment = ArchiveType("http://linked.earth/ontology/archive#LakeSediment", "Lake sediment")
    MarineSediment = ArchiveType("http://linked.earth/ontology/archive#MarineSediment", "Marine sediment")
    Midden = ArchiveType("http://linked.earth/ontology/archive#Midden", "Midden")
    MolluskShell = ArchiveType("http://linked.earth/ontology/archive#MolluskShell", "Mollusk shell")
    Peat = ArchiveType("http://linked.earth/ontology/archive#Peat", "Peat")
    Sclerosponge = ArchiveType("http://linked.earth/ontology/archive#Sclerosponge", "Sclerosponge")
    Shoreline = ArchiveType("http://linked.earth/ontology/archive#Shoreline", "Shoreline")
    Speleothem = ArchiveType("http://linked.earth/ontology/archive#Speleothem", "Speleothem")
    TerrestrialSediment = ArchiveType("http://linked.earth/ontology/archive#TerrestrialSediment", "Terrestrial sediment")
    Wood = ArchiveType("http://linked.earth/ontology/archive#Wood", "Wood")
    Documents = ArchiveType("http://linked.earth/ontology/archive#Documents", "Documents")
    Other = ArchiveType("http://linked.earth/ontology/archive#Other", "Other")



##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid
from pylipd.classes.change import Change

class ChangeLog:
    """Auto-generated LinkedEarth class representing `ChangeLog`."""
    def __init__(self):
        """Initialize a new ChangeLog instance."""
        
        self.changess: list[Change] = []
        self.curator: str = None
        self.lastVersion: str = None
        self.notes: str = None
        self.timestamp: str = None
        self.version: str = None
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#ChangeLog"
        self.id = self.ns + "/" + uniqid("ChangeLog.")

    @staticmethod
    def from_data(id, data) -> 'ChangeLog':
        """Instantiate `ChangeLog` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        ChangeLog
            The populated `ChangeLog` instance.
        """
        self = ChangeLog()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasChanges":
                for val in value:
                    if "@id" in val:
                        obj = Change.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.changess.append(obj)
        
            elif key == "hasCurator":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.curator = obj
        
            elif key == "hasLastVersion":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.lastVersion = obj
        
            elif key == "hasNotes":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.notes = obj
        
            elif key == "hasTimestamp":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.timestamp = obj
        
            elif key == "hasVersion":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.version = obj
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if len(self.changess):
            data[self.id]["hasChanges"] = []
        for value_obj in self.changess:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasChanges"].append(obj)

        if self.curator:
            value_obj = self.curator
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasCurator"] = [obj]
                

        if self.lastVersion:
            value_obj = self.lastVersion
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasLastVersion"] = [obj]
                

        if self.notes:
            value_obj = self.notes
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasNotes"] = [obj]
                

        if self.timestamp:
            value_obj = self.timestamp
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasTimestamp"] = [obj]
                

        if self.version:
            value_obj = self.version
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasVersion"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if len(self.changess):
            data["changes"] = []
        for value_obj in self.changess:
            obj = value_obj.to_json()
            data["changes"].append(obj)

        if self.curator:
            value_obj = self.curator
            obj = value_obj
            data["curator"] = obj

        if self.lastVersion:
            value_obj = self.lastVersion
            obj = value_obj
            data["lastVersion"] = obj

        if self.notes:
            value_obj = self.notes
            obj = value_obj
            data["notes"] = obj

        if self.timestamp:
            value_obj = self.timestamp
            obj = value_obj
            data["timestamp"] = obj

        if self.version:
            value_obj = self.version
            obj = value_obj
            data["version"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'ChangeLog':
        """Instantiate `ChangeLog` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        ChangeLog
            The populated `ChangeLog` instance.
        """
        self = ChangeLog()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "changes":
                for value in pvalue:
                    obj = Change.from_json(value)
                    self.changess.append(obj)
            elif key == "curator":
                    value = pvalue
                    obj = value
                    self.curator = obj
            elif key == "lastVersion":
                    value = pvalue
                    obj = value
                    self.lastVersion = obj
            elif key == "notes":
                    value = pvalue
                    obj = value
                    self.notes = obj
            elif key == "timestamp":
                    value = pvalue
                    obj = value
                    self.timestamp = obj
            elif key == "version":
                    value = pvalue
                    obj = value
                    self.version = obj
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getChangess(self) -> list[Change]:
        """Get changess list.

        Returns
        -------
        list[Change]
            A list of Change objects.
        """
        return self.changess

    def setChangess(self, changess:list[Change]):
        """Set the changess list.

        Parameters
        ----------
        changess : list[Change]
            The list to assign.
        """
        assert isinstance(changess, list), "Error: changess is not a list"
        assert all(isinstance(x, Change) for x in changess), f"Error: '{changess}' is not of type Change"
        self.changess = changess

    def addChanges(self, changess:Change):
        """Add a value to the changess list.

        Parameters
        ----------
        changess : Change
            The value to append.
        """
        assert isinstance(changess, Change), f"Error: '{changess}' is not of type Change"
        self.changess.append(changess)
        
    def getCurator(self) -> str:
        """Get curator.

        Returns
        -------
        str
            The current value of curator.
        """
        return self.curator

    def setCurator(self, curator:str):
        """Set curator.

        Parameters
        ----------
        curator : str
            The value to assign.
        """
        assert isinstance(curator, str), f"Error: '{curator}' is not of type str"
        self.curator = curator
    
    def getLastVersion(self) -> str:
        """Get lastVersion.

        Returns
        -------
        str
            The current value of lastVersion.
        """
        return self.lastVersion

    def setLastVersion(self, lastVersion:str):
        """Set lastVersion.

        Parameters
        ----------
        lastVersion : str
            The value to assign.
        """
        assert isinstance(lastVersion, str), f"Error: '{lastVersion}' is not of type str"
        self.lastVersion = lastVersion
    
    def getNotes(self) -> str:
        """Get notes.

        Returns
        -------
        str
            The current value of notes.
        """
        return self.notes

    def setNotes(self, notes:str):
        """Set notes.

        Parameters
        ----------
        notes : str
            The value to assign.
        """
        assert isinstance(notes, str), f"Error: '{notes}' is not of type str"
        self.notes = notes
    
    def getTimestamp(self) -> str:
        """Get timestamp.

        Returns
        -------
        str
            The current value of timestamp.
        """
        return self.timestamp

    def setTimestamp(self, timestamp:str):
        """Set timestamp.

        Parameters
        ----------
        timestamp : str
            The value to assign.
        """
        assert isinstance(timestamp, str), f"Error: '{timestamp}' is not of type str"
        self.timestamp = timestamp
    
    def getVersion(self) -> str:
        """Get version.

        Returns
        -------
        str
            The current value of version.
        """
        return self.version

    def setVersion(self, version:str):
        """Set version.

        Parameters
        ----------
        version : str
            The value to assign.
        """
        assert isinstance(version, str), f"Error: '{version}' is not of type str"
        self.version = version
    


from rdflib import RDF, RDFS
from rdflib.graph import URIRef, Literal
from rdflib.namespace import XSD

from ..globals.urls import ONTONS

class JSONToRDF:
    def __init__(self, graph, graphurl):
        self.graph = graph
        self.graphurl = graphurl

    def _load_triple_into_graph(self, subject, prop, value):
        for val in value:
            valitem = None
            if val["@type"] == "uri":
                valitem = URIRef(val["@id"])
            elif val["@type"] == "literal":
                dtype:str = val["@datatype"]
                if dtype:
                    if type(val["@value"]) == str:
                        dtype="http://www.w3.org/2001/XMLSchema#string"
                    valitem = Literal(val["@value"], datatype=URIRef(dtype))
                else:
                    valitem = Literal(val["@value"])
            if valitem:
                self.graph.add((
                    URIRef(subject),
                    prop,
                    valitem,
                    URIRef(self.graphurl)
                ))

    def _clear_subgraph(self):
        for ctx in self.graph.contexts():
            id = str(ctx.identifier)
            if id == self.graphurl:
                self.graph.remove((None, None, None, URIRef(self.graphurl)))    

    def load_data_in_graph(self, data):        
        # Clear the subgraph
        self._clear_subgraph()

        # Load data
        for subject in data:
            for propid in data[subject]:
                value = data[subject][propid]
                prop = URIRef(ONTONS + propid)
                if propid == "type":
                    prop = RDF.type
                if propid == "label":
                    prop = RDFS.label
                self._load_triple_into_graph(subject, prop, value)




##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid
from pylipd.classes.person import Person

class Funding:
    """Auto-generated LinkedEarth class representing `Funding`."""
    def __init__(self):
        """Initialize a new Funding instance."""
        
        self.fundingAgency: str = None
        self.fundingCountry: str = None
        self.grants: list[str] = []
        self.investigators: list[Person] = []
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#Funding"
        self.id = self.ns + "/" + uniqid("Funding.")

    @staticmethod
    def from_data(id, data) -> 'Funding':
        """Instantiate `Funding` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        Funding
            The populated `Funding` instance.
        """
        self = Funding()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasFundingAgency":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.fundingAgency = obj
        
            elif key == "hasFundingCountry":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.fundingCountry = obj
        
            elif key == "hasGrant":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]
                    self.grants.append(obj)
        
            elif key == "hasInvestigator":
                for val in value:
                    if "@id" in val:
                        obj = Person.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.investigators.append(obj)
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if len(self.grants):
            data[self.id]["hasGrant"] = []
        for value_obj in self.grants:
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasGrant"].append(obj)

        if len(self.investigators):
            data[self.id]["hasInvestigator"] = []
        for value_obj in self.investigators:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasInvestigator"].append(obj)

        if self.fundingAgency:
            value_obj = self.fundingAgency
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasFundingAgency"] = [obj]
                

        if self.fundingCountry:
            value_obj = self.fundingCountry
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasFundingCountry"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if len(self.grants):
            data["grant"] = []
        for value_obj in self.grants:
            obj = value_obj
            data["grant"].append(obj)

        if len(self.investigators):
            data["investigator"] = []
        for value_obj in self.investigators:
            obj = value_obj.to_json()
            data["investigator"].append(obj)

        if self.fundingAgency:
            value_obj = self.fundingAgency
            obj = value_obj
            data["agency"] = obj

        if self.fundingCountry:
            value_obj = self.fundingCountry
            obj = value_obj
            data["country"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'Funding':
        """Instantiate `Funding` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        Funding
            The populated `Funding` instance.
        """
        self = Funding()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "agency":
                    value = pvalue
                    obj = value
                    self.fundingAgency = obj
            elif key == "country":
                    value = pvalue
                    obj = value
                    self.fundingCountry = obj
            elif key == "grant":
                for value in pvalue:
                    obj = value
                    self.grants.append(obj)
            elif key == "investigator":
                for value in pvalue:
                    obj = Person.from_json(value)
                    self.investigators.append(obj)
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getFundingAgency(self) -> str:
        """Get fundingAgency.

        Returns
        -------
        str
            The current value of fundingAgency.
        """
        return self.fundingAgency

    def setFundingAgency(self, fundingAgency:str):
        """Set fundingAgency.

        Parameters
        ----------
        fundingAgency : str
            The value to assign.
        """
        assert isinstance(fundingAgency, str), f"Error: '{fundingAgency}' is not of type str"
        self.fundingAgency = fundingAgency
    
    def getFundingCountry(self) -> str:
        """Get fundingCountry.

        Returns
        -------
        str
            The current value of fundingCountry.
        """
        return self.fundingCountry

    def setFundingCountry(self, fundingCountry:str):
        """Set fundingCountry.

        Parameters
        ----------
        fundingCountry : str
            The value to assign.
        """
        assert isinstance(fundingCountry, str), f"Error: '{fundingCountry}' is not of type str"
        self.fundingCountry = fundingCountry
    
    def getGrants(self) -> list[str]:
        """Get grants list.

        Returns
        -------
        list[str]
            A list of str objects.
        """
        return self.grants

    def setGrants(self, grants:list[str]):
        """Set the grants list.

        Parameters
        ----------
        grants : list[str]
            The list to assign.
        """
        assert isinstance(grants, list), "Error: grants is not a list"
        assert all(isinstance(x, str) for x in grants), f"Error: '{grants}' is not of type str"
        self.grants = grants

    def addGrant(self, grants:str):
        """Add a value to the grants list.

        Parameters
        ----------
        grants : str
            The value to append.
        """
        assert isinstance(grants, str), f"Error: '{grants}' is not of type str"
        self.grants.append(grants)
        
    def getInvestigators(self) -> list[Person]:
        """Get investigators list.

        Returns
        -------
        list[Person]
            A list of Person objects.
        """
        return self.investigators

    def setInvestigators(self, investigators:list[Person]):
        """Set the investigators list.

        Parameters
        ----------
        investigators : list[Person]
            The list to assign.
        """
        assert isinstance(investigators, list), "Error: investigators is not a list"
        assert all(isinstance(x, Person) for x in investigators), f"Error: '{investigators}' is not of type Person"
        self.investigators = investigators

    def addInvestigator(self, investigators:Person):
        """Add a value to the investigators list.

        Parameters
        ----------
        investigators : Person
            The value to append.
        """
        assert isinstance(investigators, Person), f"Error: '{investigators}' is not of type Person"
        self.investigators.append(investigators)
        



##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid

class Calibration:
    """Auto-generated LinkedEarth class representing `Calibration`."""
    def __init__(self):
        """Initialize a new Calibration instance."""
        
        self.dOI: str = None
        self.datasetRange: str = None
        self.equation: str = None
        self.equationIntercept: str = None
        self.equationR2: str = None
        self.equationSlope: str = None
        self.equationSlopeUncertainty: str = None
        self.method: str = None
        self.methodDetail: str = None
        self.notes: str = None
        self.proxyDataset: str = None
        self.seasonality: str = None
        self.targetDataset: str = None
        self.uncertainty: str = None
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#Calibration"
        self.id = self.ns + "/" + uniqid("Calibration.")

    @staticmethod
    def from_data(id, data) -> 'Calibration':
        """Instantiate `Calibration` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        Calibration
            The populated `Calibration` instance.
        """
        self = Calibration()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasDOI":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.dOI = obj
        
            elif key == "hasDatasetRange":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.datasetRange = obj
        
            elif key == "hasEquation":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.equation = obj
        
            elif key == "hasEquationIntercept":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.equationIntercept = obj
        
            elif key == "hasEquationR2":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.equationR2 = obj
        
            elif key == "hasEquationSlope":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.equationSlope = obj
        
            elif key == "hasEquationSlopeUncertainty":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.equationSlopeUncertainty = obj
        
            elif key == "hasMethod":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.method = obj
        
            elif key == "hasMethodDetail":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.methodDetail = obj
        
            elif key == "hasNotes":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.notes = obj
        
            elif key == "hasProxyDataset":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.proxyDataset = obj
        
            elif key == "hasTargetDataset":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.targetDataset = obj
        
            elif key == "hasUncertainty":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.uncertainty = obj
        
            elif key == "seasonality":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.seasonality = obj
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if self.dOI:
            value_obj = self.dOI
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasDOI"] = [obj]
                

        if self.datasetRange:
            value_obj = self.datasetRange
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasDatasetRange"] = [obj]
                

        if self.equation:
            value_obj = self.equation
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasEquation"] = [obj]
                

        if self.equationIntercept:
            value_obj = self.equationIntercept
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasEquationIntercept"] = [obj]
                

        if self.equationR2:
            value_obj = self.equationR2
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasEquationR2"] = [obj]
                

        if self.equationSlope:
            value_obj = self.equationSlope
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasEquationSlope"] = [obj]
                

        if self.equationSlopeUncertainty:
            value_obj = self.equationSlopeUncertainty
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasEquationSlopeUncertainty"] = [obj]
                

        if self.method:
            value_obj = self.method
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasMethod"] = [obj]
                

        if self.methodDetail:
            value_obj = self.methodDetail
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasMethodDetail"] = [obj]
                

        if self.notes:
            value_obj = self.notes
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasNotes"] = [obj]
                

        if self.proxyDataset:
            value_obj = self.proxyDataset
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasProxyDataset"] = [obj]
                

        if self.seasonality:
            value_obj = self.seasonality
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["seasonality"] = [obj]
                

        if self.targetDataset:
            value_obj = self.targetDataset
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasTargetDataset"] = [obj]
                

        if self.uncertainty:
            value_obj = self.uncertainty
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasUncertainty"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if self.dOI:
            value_obj = self.dOI
            obj = value_obj
            data["doi"] = obj

        if self.datasetRange:
            value_obj = self.datasetRange
            obj = value_obj
            data["datasetRange"] = obj

        if self.equation:
            value_obj = self.equation
            obj = value_obj
            data["equation"] = obj

        if self.equationIntercept:
            value_obj = self.equationIntercept
            obj = value_obj
            data["equationIntercept"] = obj

        if self.equationR2:
            value_obj = self.equationR2
            obj = value_obj
            data["equationR2"] = obj

        if self.equationSlope:
            value_obj = self.equationSlope
            obj = value_obj
            data["equationSlope"] = obj

        if self.equationSlopeUncertainty:
            value_obj = self.equationSlopeUncertainty
            obj = value_obj
            data["equationSlopeUncertainty"] = obj

        if self.method:
            value_obj = self.method
            obj = value_obj
            data["method"] = obj

        if self.methodDetail:
            value_obj = self.methodDetail
            obj = value_obj
            data["methodDetail"] = obj

        if self.notes:
            value_obj = self.notes
            obj = value_obj
            data["notes"] = obj

        if self.proxyDataset:
            value_obj = self.proxyDataset
            obj = value_obj
            data["proxyDataset"] = obj

        if self.seasonality:
            value_obj = self.seasonality
            obj = value_obj
            data["hasSeasonality"] = obj

        if self.targetDataset:
            value_obj = self.targetDataset
            obj = value_obj
            data["targetDataset"] = obj

        if self.uncertainty:
            value_obj = self.uncertainty
            obj = value_obj
            data["uncertainty"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'Calibration':
        """Instantiate `Calibration` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        Calibration
            The populated `Calibration` instance.
        """
        self = Calibration()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "datasetRange":
                    value = pvalue
                    obj = value
                    self.datasetRange = obj
            elif key == "doi":
                    value = pvalue
                    obj = value
                    self.dOI = obj
            elif key == "equation":
                    value = pvalue
                    obj = value
                    self.equation = obj
            elif key == "equationIntercept":
                    value = pvalue
                    obj = value
                    self.equationIntercept = obj
            elif key == "equationR2":
                    value = pvalue
                    obj = value
                    self.equationR2 = obj
            elif key == "equationSlope":
                    value = pvalue
                    obj = value
                    self.equationSlope = obj
            elif key == "equationSlopeUncertainty":
                    value = pvalue
                    obj = value
                    self.equationSlopeUncertainty = obj
            elif key == "hasSeasonality":
                    value = pvalue
                    obj = value
                    self.seasonality = obj
            elif key == "method":
                    value = pvalue
                    obj = value
                    self.method = obj
            elif key == "methodDetail":
                    value = pvalue
                    obj = value
                    self.methodDetail = obj
            elif key == "notes":
                    value = pvalue
                    obj = value
                    self.notes = obj
            elif key == "proxyDataset":
                    value = pvalue
                    obj = value
                    self.proxyDataset = obj
            elif key == "targetDataset":
                    value = pvalue
                    obj = value
                    self.targetDataset = obj
            elif key == "uncertainty":
                    value = pvalue
                    obj = value
                    self.uncertainty = obj
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getDOI(self) -> str:
        """Get dOI.

        Returns
        -------
        str
            The current value of dOI.
        """
        return self.dOI

    def setDOI(self, dOI:str):
        """Set dOI.

        Parameters
        ----------
        dOI : str
            The value to assign.
        """
        assert isinstance(dOI, str), f"Error: '{dOI}' is not of type str"
        self.dOI = dOI
    
    def getDatasetRange(self) -> str:
        """Get datasetRange.

        Returns
        -------
        str
            The current value of datasetRange.
        """
        return self.datasetRange

    def setDatasetRange(self, datasetRange:str):
        """Set datasetRange.

        Parameters
        ----------
        datasetRange : str
            The value to assign.
        """
        assert isinstance(datasetRange, str), f"Error: '{datasetRange}' is not of type str"
        self.datasetRange = datasetRange
    
    def getEquation(self) -> str:
        """Get equation.

        Returns
        -------
        str
            The current value of equation.
        """
        return self.equation

    def setEquation(self, equation:str):
        """Set equation.

        Parameters
        ----------
        equation : str
            The value to assign.
        """
        assert isinstance(equation, str), f"Error: '{equation}' is not of type str"
        self.equation = equation
    
    def getEquationIntercept(self) -> str:
        """Get equationIntercept.

        Returns
        -------
        str
            The current value of equationIntercept.
        """
        return self.equationIntercept

    def setEquationIntercept(self, equationIntercept:str):
        """Set equationIntercept.

        Parameters
        ----------
        equationIntercept : str
            The value to assign.
        """
        assert isinstance(equationIntercept, str), f"Error: '{equationIntercept}' is not of type str"
        self.equationIntercept = equationIntercept
    
    def getEquationR2(self) -> str:
        """Get equationR2.

        Returns
        -------
        str
            The current value of equationR2.
        """
        return self.equationR2

    def setEquationR2(self, equationR2:str):
        """Set equationR2.

        Parameters
        ----------
        equationR2 : str
            The value to assign.
        """
        assert isinstance(equationR2, str), f"Error: '{equationR2}' is not of type str"
        self.equationR2 = equationR2
    
    def getEquationSlope(self) -> str:
        """Get equationSlope.

        Returns
        -------
        str
            The current value of equationSlope.
        """
        return self.equationSlope

    def setEquationSlope(self, equationSlope:str):
        """Set equationSlope.

        Parameters
        ----------
        equationSlope : str
            The value to assign.
        """
        assert isinstance(equationSlope, str), f"Error: '{equationSlope}' is not of type str"
        self.equationSlope = equationSlope
    
    def getEquationSlopeUncertainty(self) -> str:
        """Get equationSlopeUncertainty.

        Returns
        -------
        str
            The current value of equationSlopeUncertainty.
        """
        return self.equationSlopeUncertainty

    def setEquationSlopeUncertainty(self, equationSlopeUncertainty:str):
        """Set equationSlopeUncertainty.

        Parameters
        ----------
        equationSlopeUncertainty : str
            The value to assign.
        """
        assert isinstance(equationSlopeUncertainty, str), f"Error: '{equationSlopeUncertainty}' is not of type str"
        self.equationSlopeUncertainty = equationSlopeUncertainty
    
    def getMethod(self) -> str:
        """Get method.

        Returns
        -------
        str
            The current value of method.
        """
        return self.method

    def setMethod(self, method:str):
        """Set method.

        Parameters
        ----------
        method : str
            The value to assign.
        """
        assert isinstance(method, str), f"Error: '{method}' is not of type str"
        self.method = method
    
    def getMethodDetail(self) -> str:
        """Get methodDetail.

        Returns
        -------
        str
            The current value of methodDetail.
        """
        return self.methodDetail

    def setMethodDetail(self, methodDetail:str):
        """Set methodDetail.

        Parameters
        ----------
        methodDetail : str
            The value to assign.
        """
        assert isinstance(methodDetail, str), f"Error: '{methodDetail}' is not of type str"
        self.methodDetail = methodDetail
    
    def getNotes(self) -> str:
        """Get notes.

        Returns
        -------
        str
            The current value of notes.
        """
        return self.notes

    def setNotes(self, notes:str):
        """Set notes.

        Parameters
        ----------
        notes : str
            The value to assign.
        """
        assert isinstance(notes, str), f"Error: '{notes}' is not of type str"
        self.notes = notes
    
    def getProxyDataset(self) -> str:
        """Get proxyDataset.

        Returns
        -------
        str
            The current value of proxyDataset.
        """
        return self.proxyDataset

    def setProxyDataset(self, proxyDataset:str):
        """Set proxyDataset.

        Parameters
        ----------
        proxyDataset : str
            The value to assign.
        """
        assert isinstance(proxyDataset, str), f"Error: '{proxyDataset}' is not of type str"
        self.proxyDataset = proxyDataset
    
    def getSeasonality(self) -> str:
        """Get seasonality.

        Returns
        -------
        str
            The current value of seasonality.
        """
        return self.seasonality

    def setSeasonality(self, seasonality:str):
        """Set seasonality.

        Parameters
        ----------
        seasonality : str
            The value to assign.
        """
        assert isinstance(seasonality, str), f"Error: '{seasonality}' is not of type str"
        self.seasonality = seasonality
    
    def getTargetDataset(self) -> str:
        """Get targetDataset.

        Returns
        -------
        str
            The current value of targetDataset.
        """
        return self.targetDataset

    def setTargetDataset(self, targetDataset:str):
        """Set targetDataset.

        Parameters
        ----------
        targetDataset : str
            The value to assign.
        """
        assert isinstance(targetDataset, str), f"Error: '{targetDataset}' is not of type str"
        self.targetDataset = targetDataset
    
    def getUncertainty(self) -> str:
        """Get uncertainty.

        Returns
        -------
        str
            The current value of uncertainty.
        """
        return self.uncertainty

    def setUncertainty(self, uncertainty:str):
        """Set uncertainty.

        Parameters
        ----------
        uncertainty : str
            The value to assign.
        """
        assert isinstance(uncertainty, str), f"Error: '{uncertainty}' is not of type str"
        self.uncertainty = uncertainty
    


if __name__ == "__main__":
    from pylipd.lipd import LiPD

    local_lipd_dir = "/Users/varun/git/LiPD/PyLiPD/data/lpd"

    L = LiPD()

    # Convert LiPD files to RDF    
    L.convert_lipd_dir_to_rdf(
        local_lipd_dir,
        local_lipd_dir+".nq", 
        parallel=True,
        standardize=True,
        add_labels=False)
    exit()
    

    #L.load_from_dir(local_lipd_dir, parallel=True, cutoff=100)
    
    print(f"Total number of datasets: {len(L.get_all_dataset_names())}")

    print("Filtering by archive type and bounding box")
    
    #Lfiltered = L.filter_by_archive_type("marine").filter_by_geo_bbox(-50, -50, 50, 50)
    
    #print(f"Total number of filtered datasets: {len(Lfiltered.get_all_dataset_names())}")

    # Free up memory by deleting the original LiPD object
    #del L

    # Convert to LiPD Series
    print("Creating LiPD Series...")
    S = L.to_lipd_series(parallel=True)

    df = S.get_timeseries_essentials()
    print(df)

    #print("Filtering LiPD Series with variable name starting with 'd18O'")
    #S = S.filter_by_name("d18O")
    
    print("Printing all valid variables")
    print(S.get_all_variables())

    exit()



from rdflib.graph import ConjunctiveGraph, URIRef
import json

class RDFToJSON:
    """
    The RDFToFacts class helps in converting an RDF Graph to Plain JSON
    It uses the SCHEMA dictionary (from globals/schema.py) to do the conversion
    """    
    def __init__(self, id, graph: ConjunctiveGraph):
        self.graph = graph
        self.allfacts = {}
        self._get_indexed_facts(id)
    
    def _get_prop_values_from_query_result_p_o(self, qres):
        result = {}
        for row in qres:
            pname = self._local_name(row.p)
            if pname not in result:
                result[pname] = []
            value = {}
            if isinstance(row.o, URIRef):
                value["@type"] = "uri"
                value["@id"] = str(row.o)
            else:
                value["@type"] = "literal"
                value["@value"] = row.o.value
                value["@datatype"] = row.o.datatype
            result[pname].append(value)
        return result

    def _get_facts(self, id):
        query = f"SELECT ?p ?o WHERE {{ <{id}> ?p ?o }}"
        qres = self.graph.query(query)
        return self._get_prop_values_from_query_result_p_o(qres)

    def _local_name(self, url):
        return str(url).split("#")[-1]
    
    def _get_indexed_facts(self, id):
        if id in self.allfacts:
            return

        facts = self._get_facts(id)
        self.allfacts[id] = facts

        for pname, pfacts in facts.items():
            for pfact in pfacts:
                if pfact["@type"] == "uri":
                    if pname != "type":
                        self._get_indexed_facts(pfact["@id"])

    def to_json(self):
        return json.dumps(self.allfacts, default=str, indent=3)